import path, { normalize, join, resolve } from 'pathe';
import { defineNuxtModule, createResolver, logger as logger$1, addTemplate, addPluginTemplate, addImports } from '@nuxt/kit';
import fs from 'fs';
import { pascalCase, paramCase } from 'change-case';
import consola from 'consola';
import clone from 'clone';
import template from 'lodash.template';
import { build as build$2, defineConfig } from 'vite';
import vuePlugin from '@vitejs/plugin-vue';

const MODULE_NAME = 'nuxt-custom-elements';
const logger = consola.withTag(MODULE_NAME);

const BUILD_DIR = 'dist';
const ENTRIES_DIR = 'entries';

function getModuleDir(nuxt) {
  return path.resolve(nuxt.options.buildDir, MODULE_NAME);
}
function getEntriesDir(nuxt) {
  return path.resolve(getModuleDir(nuxt), ENTRIES_DIR);
}
function getBuildDir(nuxt) {
  return path.resolve(getModuleDir(nuxt), BUILD_DIR);
}
function getDistDir(nuxt) {
  return path.resolve(nuxt._nitro.options.output.publicDir, MODULE_NAME);
}

const getDefaultOptions = () => {
  return {
    name: MODULE_NAME,
    buildDir: null,
    analyzer: false,
    entries: []
  };
};

/**
 * Prepare entry and its tags.
 * @param {Object} entry Endpoint-Entry
 */
function prepareEntry(entry) {
  const tags = entry.tags.map(tag => {
    return Object.assign(tag, {
      name: paramCase(tag.name),
      options: tag.options,
      async: tag.async || false
    });
  });
  return {
    ...entry,
    name: paramCase(entry.name),
    tags
  };
}

/**
 * Create prepared entry list.
 * @param {Object} nuxt Nuxt
 * @param {Object} options Module-Options
 */
function generateEntries(runtimeDir, nuxt, moduleOptions) {
  return moduleOptions.entries.map(entry => {
    entry = prepareEntry(entry);
    return {
      name: entry.name,
      template: Object.fromEntries(
        ['client'].map(type => {
          return [
            type,
            {
              src: path.resolve(runtimeDir, 'tmpl', 'entry.mjs'),
              options: {
                tags: entry.tags,
                // TODO: Hier muss auhc noch ist der dann auch für vite?
                webpackExtend: entry.webpackExtend
              },
              fileName: path.resolve(
                getEntriesDir(nuxt),
                `${entry.name}.${type}.mjs`
              )
            }
          ];
        })
      )
    };
  });
}

async function copyBuild(from, to) {
  try {
    await fs.promises.rm(to, { recursive: true, force: true });
  } catch (error) {
    // directory not found
  }
  try {
    await fs.promises.cp(from, to, { recursive: true });
    logger.info(`Entries output directory: \`${to}\``);
    logger.success('Generated entries!');
  } catch (error) {
    logger.warn('Generated entries `dist` is missing!');
  }
}

async function onClose(nuxt, options) {
  const buildDir = getBuildDir(nuxt);
  let distPath = getDistDir(nuxt);

  // Clean destination folder
  if (options.buildDir) {
    distPath = options.buildDir;
  }

  await copyBuild(buildDir, distPath);
}

function getEntryNamingMap(options) {
  return options.entries.reduce((result, { name }) => {
    result[String(pascalCase(name))] = paramCase(name);
    result[String(paramCase(name))] = paramCase(name);
    return result;
  }, {});
}

function DEFAULT_FILENAME_FUNC(webpackConfig, moduleOptions) {
  if (moduleOptions.modern) {
    if (webpackConfig.name === 'modern') {
      return '[name].modern.js';
    } else {
      return '[name].client.js';
    }
  } else {
    return '[name].js';
  }
}

function DEFAULT_CHUNK_FILENAME_FUNC(webpackConfig, moduleOptions) {
  if (moduleOptions.modern) {
    return '[name].[hash].js';
  } else {
    return '[name].js';
  }
}
function getDefaultWebpackOutputOptions() {
  return {
    filename: DEFAULT_FILENAME_FUNC,
    chunkFilename: DEFAULT_CHUNK_FILENAME_FUNC,
    publicPath: './'
  };
}

function getTagsFromEntry(entry) {
  const tags = [];
  if ('tags' in entry) {
    tags.push(...entry.tags);
  } else {
    tags.push(entry);
  }
  return tags;
}

function getTagHTMLFromEntry(entry) {
  return getTagsFromEntry(entry).map(tag => {
    const options =
      typeof tag.options === 'function' ? tag.options() : tag.options || {};

    let props = [];
    if ('props' in options) {
      if (Array.isArray(options.props) && options.props.length > 0) {
        // array ['prop-a', prop-b]
        props = options.props.map(prop => `${paramCase(prop)}=""`);
      } else {
        // object {'prop-a': 'val-a', 'prop-b': 'val-a'}
        props = Object.keys(options.props || {}).reduce((result, prop) => {
          const value = options.props[String(prop)];
          if (typeof value !== 'object') {
            result.push(`${paramCase(prop)}="${value}"`);
          }
          return result;
        }, []);
      }
    }

    const slotContent = tag.slotContent || '';
    return `<${tag.name} ${props.join(' ')}>${slotContent}</${tag.name}>`;
  });
}

async function webpackBuild(config) {
  const Webpack = await import('webpack').then(module => module.default);
  return new Promise(resolve => {
    Webpack(config, (err, stats) => {
      if (err || stats.hasErrors()) {
        throw err;
      }
      resolve(stats);
    });
  });
}

async function build$1(webpackConfigs, statsList = []) {
  const configs = webpackConfigs.splice(0, Math.min(1, webpackConfigs.length));
  if (configs.length > 0) {
    const stats = await Promise.all(
      configs.map(async config => ({
        name: Object.keys(config.entry)[0],
        stats: await webpackBuild(config)
      }))
    );
    return build$1(webpackConfigs, statsList.concat(stats));
  } else {
    return statsList;
  }
}

async function getWebpackConfig(runtimeDir, entryName, nuxt, config, options) {
  const buildDir = path.normalize(path.join(getBuildDir(nuxt), entryName));
  const pluginExcludes = [
    'VueSSRClientPlugin',
    'CorsPlugin',
    'HtmlWebpackPlugin',
    'BundleAnalyzerPlugin',
    // Remove all native Objects, PWA Modul registriert sich ohne identifizierbares Object (PWA)
    'Object'
  ];

  const htmlWebpackPluginIndex = config.plugins.indexOf(
    config.plugins.find(
      plugin => plugin.constructor.name === 'HtmlWebpackPlugin'
    )
  );

  let rules = clone(config.module.rules);
  rules = setLoaderRulesForShadowMode(rules);

  const plugins = config.plugins.filter(
    plugin => !pluginExcludes.includes(plugin.constructor.name)
  );

  plugins.splice(
    htmlWebpackPluginIndex,
    0,
    ...(await createHtmlWebpackPlugins(
      runtimeDir,
      options.entries.filter(({ name }) => entryName === paramCase(name)),
      options.publicPath
    ))
  );

  plugins.push(...(await getBundleAnalyzerPlugin(options, config, entryName)));

  const output = getDefaultWebpackOutputOptions();

  const { filename, chunkFilename } = output;
  const webpackExtend =
    options.entries.find(({ name }) => paramCase(name) === entryName)
      .webpackExtend || (config => config);
  return await webpackExtend(
    Object.assign({}, config, {
      target: 'web',
      entry: prepareEntries(config, options),
      output: Object.assign(output, {
        filename: resolveFilename(filename, config, options),
        chunkFilename: resolveFilename(chunkFilename, config, options),
        path: buildDir,
        chunkLoadingGlobal: getChunkLoadingGlobalName(entryName)
      }),
      optimization: {
        runtimeChunk: false
      },
      plugins,
      module: {
        ...config.module,
        rules
      }
    }),
    { client: true }
  );
}

function getChunkLoadingGlobalName(entryName) {
  return 'webpackChunkLoadingGlobalNuxtCustomElements' + pascalCase(entryName);
}

function resolveFilename(filename, config, options) {
  if (typeof filename === 'function') {
    return filename(config, options);
  }
  return filename;
}

function prepareEntries(config, options) {
  return Object.fromEntries(
    Object.keys(options.entry).map(key => {
      const files = config.entry.app.map(file => {
        if (file.endsWith('/entry')) {
          return options.entry[String(key)].client;
        }
        return file;
      });
      return [key, files];
    })
  );
}

async function getBundleAnalyzerPlugin(options, config, entryName) {
  const { BundleAnalyzerPlugin } = await import('webpack-bundle-analyzer');
  // BundleAnalyzerPlugin
  if (options.analyzer) {
    const analyzerOptions = Object.assign(
      {
        reportFilename: path.resolve(
          `.reports/webpack/${MODULE_NAME}/${entryName}/${config.name}.html`
        ),
        statsFilename: path.resolve(
          `.reports/webpack/${MODULE_NAME}/${entryName}/stats/${config.name}.json`
        ),
        analyzerMode: 'static',
        generateStatsFile: true,
        openAnalyzer: false,
        logLevel: 'info',
        defaultSizes: 'gzip',
        statsOptions: 'normal'
      },
      options.analyze
    );
    return [new BundleAnalyzerPlugin(analyzerOptions)];
  }
  return [];
}

async function createHtmlWebpackPlugin(
  runtimeDir,
  chunks,
  publicPath,
  filename,
  tags
) {
  const HtmlWebpackPlugin = await import('html-webpack-plugin').then(
    module => module.default
  );
  filename = filename + '.html';
  return new HtmlWebpackPlugin({
    title: chunks.join(' - '),
    chunks: 'all',
    template: path.resolve(runtimeDir, '../runtime/tmpl/webpack', 'index.html'),
    filename,
    templateParameters: (compilation, assets, assetTags) => {
      return {
        htmlWebpackPlugin: {
          tags: assetTags,
          files: assets,
          options: {
            title: assetTags.title
          }
        },
        publicPath,
        tags,
        base: path.relative(`/${path.dirname(filename)}`, '/')
      };
    }
  });
}

async function createHtmlWebpackPlugins(runtimeDir, entries, publicPath) {
  return [
    await createHtmlWebpackPlugin(
      runtimeDir,
      entries.map(entry => entry.name),
      publicPath,
      'index',
      entries.reduce((result, entry) => {
        result.push(...getTagHTMLFromEntry(entry));
        return result;
      }, [])
    )
  ];
}

function setLoaderRulesForShadowMode(rules) {
  const vueLoaderRule = rules.find(({ test }) => test.test('.vue'));

  if (vueLoaderRule) {
    vueLoaderRule.use[0].options.customElement = true;
  } else {
    throw new Error("Can't find `vue-loader`…");
  }

  return rules;
}

function prepareEntryConfigs$1(runtimeDir, webpackConfig, nuxt, options) {
  return Promise.all(
    Object.keys(options.entry).map(entryName => {
      return getWebpackConfig(
        runtimeDir,
        entryName,
        nuxt,
        webpackConfig,
        Object.assign({}, options, {
          entry: { [entryName]: options.entry[String(entryName)] }
        })
      );
    })
  );
}

class Builder {
  constructor(nuxt, moduleOptions, runtimeDir) {
    this.nuxt = nuxt;
    this.moduleOptions = moduleOptions;
    this.runtimeDir = runtimeDir;
  }
}

class WebpackBuilder extends Builder {
  constructor(...args) {
    super(...args);

    if (this.nuxt.options.dev) {
      return;
    }

    this.config = null;
    this.nuxt.hook('webpack:config', configs => {
      return (this.config = configs.find(({ name }) => name === 'client'));
    });
  }

  async build() {
    const configs = await prepareEntryConfigs$1(
      this.runtimeDir,
      this.config,
      this.nuxt,
      this.moduleOptions
    );
    return build$1(configs.flat());
  }
}

async function build(builder, entryConfigs, statsList = []) {
  const configs = entryConfigs.splice(0, Math.min(1, entryConfigs.length));
  if (configs.length > 0) {
    const stats = await Promise.all(
      configs.map(async ({ name, entry, config }) => ({
        name,
        stats: await build$2(config),
        entry
      }))
    );
    return build(builder, entryConfigs, statsList.concat(stats));
  } else {
    return statsList;
  }
}

function getViteConfig(entryName, nuxt, config, options) {
  const buildDir = normalize(join(getBuildDir(nuxt), entryName));
  const entry = options.entry[String(entryName)].client;

  config = clone(config);

  config.base = '/';

  config.build = {
    ...config.build,
    target: 'esnext',
    outDir: buildDir,
    rollupOptions: {
      ...config.rollupOptions,
      input: entry
    }
  };

  config.vue.customElement = true;
  // Replace vue plugin for property customElement
  config.plugins[
    config.plugins.indexOf(
      config.plugins.find(({ name }) => name === 'vite:vue')
    )
  ] = vuePlugin(config.vue);

  config.entry = entry;

  return defineConfig(config);
}

function prepareEntryConfigs(viteConfig, nuxt, options) {
  return Promise.all(
    Object.keys(options.entry).map(entryName => {
      return {
        name: entryName,
        entry: options.entries.find(
          ({ name }) => entryName === paramCase(name)
        ),
        config: getViteConfig(
          entryName,
          nuxt,
          viteConfig,
          Object.assign({}, options, {
            entry: { [entryName]: options.entry[String(entryName)] }
          })
        )
      };
    })
  );
}

class ViteBuilder extends Builder {
  constructor(...args) {
    super(...args);

    if (this.nuxt.options.dev) {
      return;
    }

    this.config = [];
    this.nuxt.hook('vite:extendConfig', config => {
      if (!('ssr' in config)) {
        return (this.config = config);
      }
    });
  }

  async build() {
    const configs = await prepareEntryConfigs(
      this.config,
      this.nuxt,
      this.moduleOptions
    );
    const builds = await build(this, configs);

    await Promise.all(
      builds.map(async ({ name, entry, stats }) => {
        const { fileName } = stats.output.find(({ isEntry }) => isEntry);
        // eslint-disable-next-line security/detect-non-literal-fs-filename
        const html = await fs.promises.readFile(
          resolve(this.runtimeDir, 'tmpl/vite/index.html'),
          { encoding: 'utf-8' }
        );
        const compiled = template(html);
        const filepath = resolve(getBuildDir(this.nuxt), name, 'index.html');
        const options = {
          title: entry.name,
          tags: getTagHTMLFromEntry(entry),
          entry: fileName
        };
        // eslint-disable-next-line security/detect-non-literal-fs-filename
        await fs.promises.writeFile(filepath, compiled({ options }), {
          encoding: 'utf-8'
        });
      })
    );
  }
}

const validBuilders = {
  '@nuxt/webpack-builder': WebpackBuilder,
  '@nuxt/vite-builder': ViteBuilder
};

var module = defineNuxtModule({
  meta: {
    name: 'nuxt-custom-elements',
    configKey: 'customElements',
    compatibility: {
      nuxt: '^3.0.x'
    }
  },
  defaults: getDefaultOptions(),

  setup(moduleOptions, nuxt) {
    const resolver = createResolver(import.meta.url);

    const runtimeDir = resolver.resolve('./runtime');
    nuxt.options.alias['#customElements'] = runtimeDir;
    nuxt.options.alias['#customElementsEntries'] = getEntriesDir(nuxt);
    nuxt.options.build.transpile.push(runtimeDir);

    let builder;
    if (validBuilders[nuxt.options.builder]) {
      builder = new validBuilders[nuxt.options.builder](
        nuxt,
        moduleOptions,
        runtimeDir
      );
    } else {
      logger$1.log(
        `Current builder \`${nuxt.options.builder}\` is incomaptible.`
      );
      return;
    }

    // create entry templates
    const entries = generateEntries(runtimeDir, nuxt, moduleOptions);
    moduleOptions.entry = entries.reduce((result, { name, template }) => {
      Object.keys(template).forEach(type => {
        const { dst } = addTemplate({
          ...template[String(type)],
          write: true
        });
        (result[String(name)] || (result[String(name)] = {}))[String(type)] =
          dst;
      });
      return result;
    }, {});

    addTemplate({
      src: resolve(runtimeDir, 'tmpl', 'entries.mjs'),
      fileName: join(getModuleDir(nuxt), 'entries.mjs'),
      write: true,
      options: Object.assign({
        entriesDir: getEntriesDir(nuxt),
        entryMap: getEntryNamingMap(moduleOptions)
      })
    });

    addPluginTemplate({
      src: resolve(runtimeDir, 'tmpl', 'plugin.mjs'),
      fileName: join(getModuleDir(nuxt), 'plugin.mjs'),
      write: true,
      options: Object.assign(
        {
          entriesDir: getEntriesDir(nuxt),
          entryMap: getEntryNamingMap(moduleOptions)
        },
        moduleOptions
      )
    });

    addImports({
      name: 'useCustomElements',
      as: 'useCustomElements',
      from: resolve(runtimeDir, 'composables/useCustomElements.mjs')
    });

    if (!nuxt.options.dev) {
      registerHooks(nuxt, moduleOptions, builder);
    }
  }
});

function registerHooks(nuxt, moduleOptions, builder) {
  nuxt.hook('build:done', async () => {
    await builder.build();
  });

  nuxt.hook('close', () => {
    return onClose(nuxt, moduleOptions);
  });
}

export { module as default };
